// File author is Ítalo Lima Marconato Matias
//
// Created on July 18 of 2018, at 21:12 BRT
// Last edited on March 29 of 2019, at 22:03 BRT

#define __CHICAGO_DISPLAY__

#include <chicago/alloc.h>
#include <chicago/arch.h>
#include <chicago/debug.h>
#include <chicago/display.h>
#include <chicago/mm.h>
#include <chicago/string.h>

UIntPtr DispFrameBuffer = 0;
UIntPtr DispBackBuffer = 0;
UInt8 DispBytesPerPixel = 0;
UIntPtr DispWidth = 0;
UIntPtr DispHeight = 0;
UInt32 DispProgressBar = 0;

UIntPtr DispGetFrameBuffer(Void) {
	return DispFrameBuffer;
}

UInt8 DispGetBytesPerPixel(Void) {
	return DispBytesPerPixel;
}

UIntPtr DispGetWidth(Void) {
	return DispWidth;
}

UIntPtr DispGetHeight(Void) {
	return DispHeight;
}

Void DispExtractARGB(UIntPtr c, PUInt8 a, PUInt8 r, PUInt8 g, PUInt8 b) {
	if (a != Null) {
		*a = (UInt8)((c >> 24) & 0xFF);
	}
	
	if (r != Null) {
		*r = (UInt8)((c >> 16) & 0xFF);
	}
	
	if (g != Null) {
		*g = (UInt8)((c >> 8) & 0xFF);
	}
	
	if (b != Null) {
		*b = (UInt8)(c & 0xFF);
	}
}

Void DispRefresh(Void) {
	StrCopyMemory((PUInt8)DispFrameBuffer, (PUInt8)DispBackBuffer, DispWidth * DispHeight * DispBytesPerPixel);
}

Void DispClearScreen(UIntPtr c) {
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
	UInt8 a;																													// Big endian...
	UInt8 a;
	UInt8 r;
	UInt8 g;
	UInt8 b;
	
	DispExtractARGB(c, &a, &r, &g, &b);																							// Extract the ARGB values
	
	c = a | (r << 8) | (g << 16) | (b << 24);																					// Convert to little endian
#endif
	
	if (DispBytesPerPixel == 3) {																								// Fill the screen with the color c
		StrSetMemory24((PUInt32)DispBackBuffer, c, DispWidth * DispHeight);
	} else if (DispBytesPerPixel == 4) {
		StrSetMemory32((PUInt32)DispBackBuffer, c, DispWidth * DispHeight);
	}
}

Void DispScrollScreen(UIntPtr c) {
	UIntPtr lsz = DispWidth * 16;
	
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
	UInt8 a;																													// Big endian...
	UInt8 a;
	UInt8 r;
	UInt8 g;
	UInt8 b;
	
	DispExtractARGB(c, &a, &r, &g, &b);																							// Extract the ARGB values
	
	c = a | (r << 8) | (g << 16) | (b << 24);																					// Convert to little endian
#endif
	
	if (DispBytesPerPixel == 3) {																								// Scroll the screen contents up and clear the last line!
		StrCopyMemory24((PUInt32)DispBackBuffer, (PUInt32)(DispBackBuffer + (lsz * 3)), DispWidth * DispHeight - lsz);
		StrSetMemory24((PUInt32)(DispBackBuffer + (DispWidth * DispHeight * 3) - (lsz * 3)), c, lsz);
	} else if (DispBytesPerPixel == 4) {
		StrCopyMemory32((PUInt32)DispBackBuffer, (PUInt32)(DispBackBuffer + (lsz * 4)), DispWidth * DispHeight - lsz);
		StrSetMemory32((PUInt32)(DispBackBuffer + (DispWidth * DispHeight * 4) - (lsz * 4)), c, lsz);
	}
}

Void DispPutPixel(UIntPtr x, UIntPtr y, UIntPtr c) {
	if (x >= DispWidth) {																										// Fix the x and the y if they are bigger than the screen dimensions
		x = DispWidth - 1;
	}
	
	if (y >= DispHeight) {
		y = DispHeight - 1;
	}
	
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
	UInt8 a;																													// Big endian...
	UInt8 a;
	UInt8 r;
	UInt8 g;
	UInt8 b;
	
	DispExtractARGB(c, &a, &r, &g, &b);																							// Extract the ARGB values
	
	c = a | (r << 8) | (g << 16) | (b << 24);																					// Convert to little endian
#endif
	
	if (DispBytesPerPixel == 3) {																								// Write the pixel!
		*((PUInt16)(DispBackBuffer + (y * (DispWidth * 3)) + (x * 3))) = (UInt16)c;
		*((PUInt8)(DispBackBuffer + (y * (DispWidth * 3)) + (x * 3) + 2)) = (UInt8)(c << 16);
	} else if (DispBytesPerPixel == 4) {
		*((PUInt32)(DispBackBuffer + (y * (DispWidth * 4)) + (x * 4))) = c;
	}
}

Void DispDrawLine(UIntPtr x0, UIntPtr y0, UIntPtr x1, UIntPtr y1, UIntPtr c) {
	if (x0 >= DispWidth) {																										// Fix the x and the y if they are bigger than the screen dimensions
		x0 = DispWidth - 1;																										//     Sorry, this function will not have a lot of documentation (not that my other ones are commented in a good way...)
	}
	
	if (y0 >= DispHeight) {
		y0 = DispHeight - 1;
	}
	
	if (x1 >= DispWidth) {
		x1 = DispWidth - 1;
	}
	
	if (y1 >= DispHeight) {
		y1 = DispHeight - 1;
	}
	
	IntPtr dx = x1 - x0;																										// Let's get the line direction
	IntPtr dy = y1 - y0;
	IntPtr sdx = (dx < 0) ? -1 : 1;
	IntPtr sdy = (dy < 0) ? -1 : 1;
	IntPtr px = x0;
	IntPtr py = y0;
	IntPtr x = 0;
	IntPtr y = 0;
	
	dx = sdx * dx + 1;
	dy = sdy * dy + 1;
	
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
	UInt8 a;																													// Big endian...
	UInt8 a;
	UInt8 r;
	UInt8 g;
	UInt8 b;
	
	DispExtractARGB(c, &a, &r, &g, &b);																							// Extract the ARGB values
	
	c = a | (r << 8) | (g << 16) | (b << 24);																					// Convert to little endian
#endif
	
	if (dx >= dy) {																												// This line is more horizontal?
		for (x = 0; x < dx; x++) {																								// Yes
			if (DispBytesPerPixel == 3) {
				*((PUInt16)(DispBackBuffer + (py * (DispWidth * 3)) + (px * 3))) = (UInt16)c;
				*((PUInt8)(DispBackBuffer + (py * (DispWidth * 3)) + (px * 3) + 2)) = (UInt8)(c << 16);
			} else if (DispBytesPerPixel == 4) {
				*((PUInt32)(DispBackBuffer + (py * (DispWidth * 4)) + (px * 4))) = c;
			}
			
			y += dy;
			
			if (y >= dx) {
				y -= dx;
				py += sdy;
			}
			
			px += sdx;
		}
	} else {
		for (y = 0; y < dy; y++) {																								// No, so is more vertical
			if (DispBytesPerPixel == 3) {
				*((PUInt16)(DispBackBuffer + (py * (DispWidth * 3)) + (px * 3))) = (UInt16)c;
				*((PUInt8)(DispBackBuffer + (py * (DispWidth * 3)) + (px * 3) + 2)) = (UInt8)(c << 16);
			} else if (DispBytesPerPixel == 4) {
				*((PUInt32)(DispBackBuffer + (py * (DispWidth * 4)) + (px * 4))) = c;
			}
			
			x += dx;
			
			if (x >= dy) {
				x -= dy;
				px += sdx;
			}
			
			py += sdy;
		}
	}
}

Void DispDrawRectangle(UIntPtr x, UIntPtr y, UIntPtr w, UIntPtr h, UIntPtr c) {
	DispDrawLine(x, y, x, y + h - 1, c);																						// Let's use the DispDrawLine to draw the rectangle (we just need to draw 4 lines)
	DispDrawLine(x, y, x + w - 1, y, c);
	DispDrawLine(x, y + h - 1, x + w - 1, y + h - 1, c);
	DispDrawLine(x + w - 1, y, x + w - 1, y + h - 1, c);
}

Void DispFillRectangle(UIntPtr x, UIntPtr y, UIntPtr w, UIntPtr h, UIntPtr c) {
	if (x >= DispWidth) {																										// Fix the x and the y if they are bigger than the screen dimensions
		x = DispWidth - 1;
	}
	
	if (y >= DispHeight) {
		y = DispHeight - 1;
	}
	
	if ((x + w) > DispWidth) {																									// Fix the width and height of our rectangle if they are bigger than the screen dimensions
		w = DispWidth - x;
	}
	
	if ((y + h) > DispHeight) {
		h = DispHeight - y;
	}
	
	PUInt32 fb = (PUInt32)(DispBackBuffer + (y * (DispWidth * DispBytesPerPixel)) + (x * DispBytesPerPixel));
	
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
	UInt8 a;																													// Big endian...
	UInt8 a;
	UInt8 r;
	UInt8 g;
	UInt8 b;
	
	DispExtractARGB(c, &a, &r, &g, &b);																							// Extract the ARGB values
	
	c = a | (r << 8) | (g << 16) | (b << 24);																					// Convert to little endian
#endif
	
	for (UIntPtr i = 0; i < h; i++) {																							// And fill our rectangle
		if (DispBytesPerPixel == 3) {
			StrSetMemory24(fb, c, w);
		} else if (DispBytesPerPixel == 4) {
			StrSetMemory32(fb, c, w);
		}
		
		fb += DispWidth;
	}
}

Void DispDrawRoundedRectangle(UIntPtr x, UIntPtr y, UIntPtr w, UIntPtr h, UIntPtr r, UIntPtr c) {
	if (r == 0) {																												// Radius is 0?
		DispDrawRectangle(x, y, w, h, c);																						// Yes, so just draw an rectangle
		return;
	}
	
	IntPtr f = 1 - r;
	IntPtr dfx = 1;
	IntPtr dfy = -2 * r;
	UIntPtr xx = 0;
	UIntPtr yy = r;
	
	if (w > 0) {
		w--;
	}
	
	if (h > 0) {
		h--;
	}
	
	while (xx < yy) {
		if (f >= 0) {
			yy--;
			dfy += 2;
			f += dfy;
		}
		
		xx++;
		dfx += 2;
		f += dfx;
		
		DispPutPixel(x - xx + r, y + h + yy - r, c);																			// Bottom left
		DispPutPixel(x - yy + r, y + h + xx - r, c);
		DispPutPixel(x - xx + r, y - yy + r, c);																				// Top left
		DispPutPixel(x - yy + r, y - xx + r, c);
		DispPutPixel(x + w + xx - r, y + h + yy - r, c);																		// Bottom right
		DispPutPixel(x + w + yy - r, y + h + xx - r, c);
		DispPutPixel(x + w + xx - r, y - yy + r, c);																			// Top right
		DispPutPixel(x + w + yy - r, y - xx + r, c);
	}
	
	DispDrawLine(x + r, y, x + w - r, y, c);																					// Draw the top line
	DispDrawLine(x + r, y + h, x + w - r, y + h, c);																			// The bottom one
	DispDrawLine(x, y + r, x, y + h - r, c);																					// The left one
	DispDrawLine(x + w, y + r, x + w, y + h - r, c);																			// And, the right one!
}

Void DispFillRoundedRectangle(UIntPtr x, UIntPtr y, UIntPtr w, UIntPtr h, UIntPtr r, UIntPtr c) {
	if (r == 0) {																												// Radius is 0?
		DispFillRectangle(x, y, w, h, c);																						// Yes, so just fill an rectangle
		return;
	}
	
	IntPtr f = 1 - r;
	IntPtr dfx = 1;
	IntPtr dfy = -2 * r;
	UIntPtr xx = 0;
	UIntPtr yy = r;
	
	if (w > 0) {
		w--;
	}
	
	if (h > 0) {
		h--;
	}
	
	while (xx < yy) {
		if (f >= 0) {
			yy--;
			dfy += 2;
			f += dfy;
		}
		
		xx++;
		dfx += 2;
		f += dfx;
		
		DispDrawLine(x - xx + r, y + h + yy - r, x - xx + r, y - yy + r, c);													// Left
		DispDrawLine(x - yy + r, y + h + xx - r, x - yy + r, y - xx + r, c);
		DispDrawLine(x + w + xx - r, y + h + yy - r, x + w + xx - r, y - yy + r, c);											// Right
		DispDrawLine(x + w + yy - r, y + h + xx - r, x + w + yy - r, y - xx + r, c);
	}
	
	DispFillRectangle(x + r, y, w - r * 2 + 1, h + 1, c);																		// Fill the center rectangle
	DispDrawLine(x, y + r, x, y + h - r, c);																					// The left line
	DispDrawLine(x + w, y + r, x + w, y + h - r, c);																			// And the right one
}

Void DispInit(UIntPtr w, UIntPtr h, UIntPtr bpp, UIntPtr fb) {
	if ((bpp != 3) && (bpp != 4)) {																								// We only support 24 and 32 bpp
		DbgWriteFormated("PANIC! Couldn't init the display\r\n");
		ArchHalt();																												// Halt
	}
	
	DispFrameBuffer = MemAAllocate(w * h * bpp, MM_PAGE_SIZE);																	// Alloc some virt space for the frame buffer
	DispBackBuffer = MemAllocate(w * h * bpp);
	DispBytesPerPixel = bpp;
	DispWidth = w;
	DispHeight = h;
	
	if (DispFrameBuffer == 0 || DispBackBuffer == 0) {
		DbgWriteFormated("PANIC! Couldn't init the display\r\n");																// Failed...
		ArchHalt();																												// Halt
	}
	
	for (UIntPtr i = 0; i < DispWidth * DispHeight * bpp; i += MM_PAGE_SIZE) {													// Let's map the frame buffer to the virtual memory!
		MmDereferencePage(MmGetPhys(DispFrameBuffer + i));																		// MemAAllocate allocated some phys addr as well, free it
		
		if (!MmMap(DispFrameBuffer + i, fb + i, MM_MAP_KDEF)) {
			DbgWriteFormated("PANIC! Couldn't init the display\r\n");
			ArchHalt();																											// Halt
		}
	}
	
	DispClearScreen(0xFF000000);																								// Clear the screen
	DispRefresh();
}
